/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.plugin.xmleditor;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.Parameter;
import org.ow2.util.stream.api.IStreamAccessor;
import org.ow2.util.stream.impl.delegator.zip.DelegatorZipStreamAccessor;
import org.ow2.util.stream.impl.file.SystemFileAccessor;
import org.w3c.dom.Document;

/**
 * Maven XML Editor Mojo.
 * @author Gael Lalire
 */
public abstract class AbstractXmlEditorMojo extends AbstractMojo {

    /**
     * @since 2.0.0
     */
    @Parameter(required = true)
    private EditorConfiguration[] editorConfigurations = null;

    /**
     * write the document to a writer.
     * @param document the document to write
     * @param writer the writer
     * @throws TransformerException if transform failed
     */
    public static void writeDocumentTo(final Document document, final Writer writer) throws TransformerException {
        Transformer transformer = TransformerFactory.newInstance().newTransformer();
        Properties properties = new Properties();
        properties.put("method", "xml");
        properties.put("version", "1.0");
        properties.put("encoding", "ISO-8859-1");
        properties.put("standalone", "yes");
        properties.put("indent", "yes");
        properties.put("omit-xml-declaration", "no");
        transformer.setOutputProperties(properties);
        DOMSource input = new DOMSource(document);
        Result output = new StreamResult(writer);
        transformer.transform(input, output);
    }

    public void execute() throws MojoExecutionException, MojoFailureException {
        for (EditorConfiguration editorConfiguration : editorConfigurations) {
            IStreamAccessor fileAccessor;
            String filePath = editorConfiguration.getFilePath();
            try {
                if (filePath == null) {
                    fileAccessor = new SystemFileAccessor(new File(editorConfiguration.getFileName()));
                } else {
                    String[] filePaths = filePath.split("!");

                    fileAccessor = new SystemFileAccessor(new File(filePaths[0]));

                    int length = filePaths.length;
                    for (int i = 1; i < length; i++) {
                        fileAccessor = new DelegatorZipStreamAccessor(fileAccessor, filePaths[i]);
                    }

                    fileAccessor = new DelegatorZipStreamAccessor(fileAccessor, editorConfiguration.getFileName());
                }
            } catch (IOException e) {
                e.printStackTrace();
                throw new MojoExecutionException("Accessing xml file issue", e);
            }

            try {
                // read xml
                DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                InputStream inputStream = fileAccessor.getInputStream();
                XMLEditor editor = new XMLEditor(documentBuilder.parse(inputStream));
                inputStream.close();

                // do replace stuff
                for (EditorOperation editorOperation : editorConfiguration.getEditorOperations()) {
                    if (editorOperation.getName().equals("set")) {
                        editor.set(editorOperation.getXmlPath(), editorOperation.getValue());
                    }
                }

                // write xml
                Writer writer = new OutputStreamWriter(fileAccessor.getOutputStream());
                writeDocumentTo(editor.getDocument(), writer);
                writer.close();

                /*
                System.out.println("On ecrit");
                writeDocumentTo(editor.getDocument(), new OutputStreamWriter(System.out));
                */
            } catch (Exception e) {
                e.printStackTrace();
                throw new MojoExecutionException("Unable to work with xml file", e);
            }
        }

    }

}
