/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.plugin.xmleditor;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * @author Gael Lalire
 */
public class XMLEditor {

    private static final String ATTRIBUTE_NAME_REGEXP = "@[\\w-]*";

    // private static final String ATTRIBUTE_NAME_CAPTURE_REGEXP = "@([\\w-]*)";

    private static final String ELEMENT_NAME_REGEXP = "[\\w-]*(?:\\[\\d*\\])?";

    /**
     * 1 : name, 2 : indice.
     */
    private static final String ELEMENT_NAME_CAPTURE_REGEXP = "([\\w-]*)(?:\\[(\\d*)\\])?";

    private static final String STRING_REGEXP = "\"[^\"]*\"";

    private static final String FILTER_REGEXP = "\\(" + ELEMENT_NAME_REGEXP + "==" + STRING_REGEXP + "\\)";

    private static final String XML_PATH_STEP_REGEXP = "(?:" + ELEMENT_NAME_REGEXP + "|" + FILTER_REGEXP + ")\\.";

    private static final String XML_PATH_LAST_STEP_REGEXP = ELEMENT_NAME_REGEXP + "|" + ATTRIBUTE_NAME_REGEXP;

    private static final String XML_PATH_REGEXP = "(?:" + XML_PATH_STEP_REGEXP + ")*(?:" + XML_PATH_LAST_STEP_REGEXP+")";

    /**
     * XML Path pattern.
     */
    private static final Pattern XML_PATH_PATTERN = Pattern.compile(XML_PATH_REGEXP);

    // capture

    private static final String XML_PATH_STEP_CAPTURE_REGEXP = "(" + ELEMENT_NAME_REGEXP + "|" + FILTER_REGEXP + ")\\.";

    /**
     * XML Path pattern.
     */
    private static final Pattern XML_PATH_STEP_CAPTURE_PATTERN = Pattern.compile(XML_PATH_STEP_CAPTURE_REGEXP);

    /**
     * Filter pattern.
     */
    private static final Pattern ELEMENT_NAME_CAPTURE_PATTERN = Pattern.compile(ELEMENT_NAME_CAPTURE_REGEXP);

    /**
     * Filter pattern.
     */
    private static final Pattern FILTER_CAPTURE_PATTERN = Pattern.compile("\\(([\\w-]*)\\[(\\d*)\\](==)(" + STRING_REGEXP
            + ")\\)");


    private static String unescapeString(final String escapedString) {
        return escapedString.substring(1, escapedString.length() - 1);
    }

    private Document document;

    public XMLEditor(final Document document) {
        this.document = document;
    }

    /**
     * @return the document
     */
    public Document getDocument() {
        return document;
    }

    protected void resolvePathStep(final List<? extends Element> initialNodeList, final String pathStep,
            final List<? super Element> resultNodeList) {
        for (Element element : initialNodeList) {
            NodeList nodeList = element.getElementsByTagName(pathStep);
            int length = nodeList.getLength();
            for (int i = 0; i < length; i++) {
                resultNodeList.add((Element) nodeList.item(i));
            }
        }
    }

    /**
     * "a.b[?>3].c[8].d[?<=2]"
     * @param initialNodeList
     * @param pathStep
     * @param number
     * @param resultNodeList
     */
    protected void resolvePathStepWithNumber(final List<? extends Element> initialNodeList, final String pathStep,
            final int number, final List<? super Element> resultNodeList) {
        for (Element element : initialNodeList) {
            NodeList nodeList = element.getElementsByTagName(pathStep);
            Element elementFiltering = (Element) nodeList.item(number);
            if (elementFiltering != null) {
                resultNodeList.add(elementFiltering);
            }
        }
    }

    /**
     * TODO: "a.(@id==n)" = { a[i] | a[i].@id = n } TODO: "a.(b[5].@id==n)" = {
     * a[i] | pour tout j -> a[i].b[5].@id = n } TODO: "a.(b.c[3].@id==n)" = {
     * a[i] | le 3e c trouve qui a pour parent un b a pour id n }
     * "a.(b[0]==text)" = { a[i] | a[i].b[0] == text } Invalid exp "a.(b==text)"
     * -> a[i].b is an element list and not a single element so we do not know
     * which element to compare. It could be an evolution that condition is ok
     * if at least one matches or if all element matches or if a parametrized
     * element matches. like this "a.(b=1=text)" "a.(b=*=text)" "a.(b=8=text)"
     * @param initialNodeList
     * @param pathStep
     * @param filterOperation
     * @param filter
     * @param resultNodeList
     */
    protected void filterPathStep(final List<? extends Element> initialNodeList, final String pathStep, final int number,
            final int filterOperation, final String filter, final List<? super Element> resultNodeList) {
        for (Element element : initialNodeList) {
            NodeList nodeList = element.getElementsByTagName(pathStep);
            Element elementFiltering = (Element) nodeList.item(number);
            if (filterOperation == 0) {
                if (filter.equals(elementFiltering.getTextContent())) {
                    resultNodeList.add(element);
                }
            }
        }
    }

    protected void setTextContent(final List<? extends Element> initialNodeList, final String value) {
        for (Element element : initialNodeList) {
            element.setTextContent(value);
        }
    }

    protected void setAttributes(final List<? extends Element> initialNodeList, final String attributeName, final String value) {
        for (Element element : initialNodeList) {
            element.setAttribute(attributeName, value);
        }
    }

    public void set(final String path, final String value) {
        Matcher matcher = XML_PATH_PATTERN.matcher(path);
        if (matcher.matches()) {
            int start = 0;
            matcher = XML_PATH_STEP_CAPTURE_PATTERN.matcher(path);
            List<Element> elementList = new ArrayList<Element>();
            List<Element> nextElementList = new ArrayList<Element>();
            elementList.add(document.getDocumentElement());

            while (matcher.find(start)) {

                String pathStep = matcher.group(1);
                if (pathStep != null) {
                    if (pathStep.startsWith("(")) {
                        Matcher filterMatcher = FILTER_CAPTURE_PATTERN.matcher(pathStep);
                        filterMatcher.matches();
                        filterPathStep(elementList, filterMatcher.group(1), Integer.parseInt(filterMatcher.group(2)), 0,
                                unescapeString(filterMatcher.group(4)), nextElementList);
                    } else {
                        Matcher elementNameMatcher = ELEMENT_NAME_CAPTURE_PATTERN.matcher(pathStep);
                        try {
                            String number = elementNameMatcher.group(2);
                            resolvePathStepWithNumber(elementList, pathStep, Integer.parseInt(number), nextElementList);
                        } catch (IllegalStateException e) {
                            resolvePathStep(elementList, pathStep, nextElementList);
                        }
                    }
                    List<Element> tmpElementList = elementList;
                    tmpElementList.clear();
                    elementList = nextElementList;
                    nextElementList = tmpElementList;
                }
                start = matcher.end();
            }
            String lastPathStep = path.substring(start);
            if (lastPathStep.startsWith("@")) {
                setAttributes(elementList, lastPathStep.substring(1), value);
            } else {
                Matcher elementNameMatcher = ELEMENT_NAME_CAPTURE_PATTERN.matcher(lastPathStep);
                try {
                    String number = elementNameMatcher.group(2);
                    resolvePathStepWithNumber(elementList, lastPathStep, Integer.parseInt(number), nextElementList);
                } catch (IllegalStateException e) {
                    resolvePathStep(elementList, lastPathStep, nextElementList);
                }
                setTextContent(nextElementList, value);
            }
        } else {
            throw new IllegalArgumentException();
        }
    }

}
